/**
 * Parametric Modeling (PMG) extension for SketchUp.
 *
 * @copyright © 2021 Samuel Tallet
 *
 * @licence GNU General Public License 3.0
 */

PMG = {}

PMG.NodesEditor = {}

PMG.NodesEditor.initializeSockets = () => {

    PMG.NodesEditor.sockets = {

        number: new Rete.Socket('数量'),
        groups: new Rete.Socket('群组'),
        point: new Rete.Socket('指向'),
        vector: new Rete.Socket('向量')

    }

}

PMG.NodesEditor.initializeSockets()

PMG.Utils = {}

PMG.Utils.cloneObject = object => {
    return JSON.parse(JSON.stringify(object))
}

PMG.Utils.isValidNumber = number => {

    if ( typeof number === 'number' ) {
        return true
    }

    if ( typeof number === 'string' && /^-?(0|[1-9][0-9]*)(\.[0-9]+)?$/.test(number) ) {
        return true
    }

    return false

}

PMG.NodesEditor.initializeControls = () => {

    PMG.NodesEditor.controls = {}

    PMG.NodesEditor.controls['number'] = {

        props: ['emitter', 'ikey', 'getData', 'putData', 'placeholder', 'readonly'],

        template: '<input type="number" :placeholder="placeholder" :title="placeholder" :readonly="readonly" :value="value" @input="change($event)" @pointerdown.stop="" @pointermove.stop="" />',
        
        data() {

            return {
                value: ''
            }

        },

        methods: {

            change(event) {

                if ( PMG.Utils.isValidNumber(event.target.value) ) {

                    this.value = parseFloat(event.target.value)

                    this.update()

                }

            },

            update() {

                this.putData(this.ikey, this.value)

                this.emitter.trigger('process')

            }

        },

        mounted() {

            if ( PMG.Utils.isValidNumber(this.getData(this.ikey)) ) {
                this.value = this.getData(this.ikey)
            }

        }
        
    }

    PMG.NodesEditor.controls['text'] = {

        props: ['emitter', 'ikey', 'getData', 'putData', 'placeholder', 'readonly'],

        template: '<input type="text" :placeholder="placeholder" :readonly="readonly" :value="value" @input="change($event)" @pointerdown.stop="" @pointermove.stop="" />',
        
        data() {

            return {
                value: ''
            }

        },

        methods: {

            change(event) {

                this.value = event.target.value

                this.update()

            },

            update() {

                this.putData(this.ikey, this.value)
                
                this.emitter.trigger('process')

            }

        },

        mounted() {
            this.value = this.getData(this.ikey)
        }
        
    }

    PMG.NodesEditor.controls['checkbox'] = {

        props: ['emitter', 'ikey', 'getData', 'putData', 'label'],

        template:
        `<div class="checkbox-control">
            <input type="checkbox" @change="change($event)" :checked="checked" />
            {{ label }}
        </div>`,

        data() {

            return {
                checked: this.getData(this.ikey),
                label: this.label
            }
            
        },

        methods: {

            change(event) {

                this.checked = event.target.checked

                this.update()

            },

            update() {

                this.putData(this.ikey, this.checked)
                
                this.emitter.trigger('process')

            }

        },

        mounted() {
            this.checked = this.getData(this.ikey)
        }
        
    }

    PMG.NodesEditor.controls['material'] = {

        props: ['emitter', 'ikey', 'getData', 'putData'],

        template:
        `<select @change="change($event)">
            <option value="">材质...</option>
            <option v-for="material in materials" v-bind:value="material.name" v-bind:selected="material.selected">
                {{ material.display_name }}
            </option>
        </select>`,

        data() {

            var nodeSelectedMaterial = this.getData(this.ikey)
            var nodeSelectableMaterials = PMG.Utils.cloneObject(SketchUpMaterials)

            nodeSelectableMaterials.forEach(nodeSelectableMaterial => {

                if ( nodeSelectableMaterial.name == nodeSelectedMaterial ) {
                    nodeSelectableMaterial.selected = 'selected'
                } else {
                    nodeSelectableMaterial.selected = ''
                }

            })

            return {
                materials: nodeSelectableMaterials
            }
            
        },

        methods: {

            change(event) {

                this.value = event.target.value

                this.update()

            },

            update() {

                this.putData(this.ikey, this.value)
                
                this.emitter.trigger('process')

            }

        },

        mounted() {
            this.value = this.getData(this.ikey)
        }
        
    }

    PMG.NodesEditor.controls['layer'] = {

        props: ['emitter', 'ikey', 'getData', 'putData'],

        template:
        `<select @change="change($event)">
            <option value="">图层/标签...</option>
            <option v-for="layer in layers" v-bind:value="layer.name" v-bind:selected="layer.selected">
                {{ layer.display_name }}
            </option>
        </select>`,

        data() {

            var nodeSelectedLayer = this.getData(this.ikey)
            var nodeSelectableLayers = PMG.Utils.cloneObject(SketchUpLayers)

            nodeSelectableLayers.forEach(nodeSelectableLayer => {

                if ( nodeSelectableLayer.name == nodeSelectedLayer ) {
                    nodeSelectableLayer.selected = 'selected'
                } else {
                    nodeSelectableLayer.selected = ''
                }

            })

            return {
                layers: nodeSelectableLayers
            }
            
        },

        methods: {

            change(event) {

                this.value = event.target.value

                this.update()

            },

            update() {

                this.putData(this.ikey, this.value)
                
                this.emitter.trigger('process')

            }

        },

        mounted() {
            this.value = this.getData(this.ikey)
        }
        
    }

}

PMG.NodesEditor.initializeControls()

class NumberReteControl extends Rete.Control {

    constructor(emitter, ikey, placeholder, readonly) {

        super(ikey)
        this.component = PMG.NodesEditor.controls.number
        this.props = { emitter, ikey, placeholder, readonly }

    }
  
    setValue(value) {
        this.vueContext.value = value
    }

}

class TextReteControl extends Rete.Control {

    constructor(emitter, ikey, placeholder, readonly) {

        super(ikey)
        this.component = PMG.NodesEditor.controls.text
        this.props = { emitter, ikey, placeholder, readonly }

    }
  
    setValue(value) {
        this.vueContext.value = value
    }

}

class CheckBoxReteControl extends Rete.Control {

    constructor(emitter, ikey, label) {

        super(ikey)
        this.component = PMG.NodesEditor.controls.checkbox
        this.props = { emitter, ikey, label }

    }
  
    setValue(checked) {
        this.vueContext.checked = checked
    }

}

class MaterialReteControl extends Rete.Control {

    constructor(emitter, key, _readonly) {

        super(key)
        this.component = PMG.NodesEditor.controls.material
        this.props = { emitter, ikey: key }

    }
  
    setValue(value) {
        this.vueContext.value = value
    }

}

class LayerReteControl extends Rete.Control {

    constructor(emitter, key, _readonly) {

        super(key)
        this.component = PMG.NodesEditor.controls.layer
        this.props = { emitter, ikey: key }

    }
  
    setValue(value) {
        this.vueContext.value = value
    }

}

class DrawBoxReteComponent extends Rete.Component {

    constructor() {
        super('绘制盒子')
    }

    builder(node) {

        var width = new Rete.Input('width', '宽度', PMG.NodesEditor.sockets.number)
        width.addControl(new NumberReteControl(this.editor, 'width', '宽度'))

        var depth = new Rete.Input('depth', '深度', PMG.NodesEditor.sockets.number)
        depth.addControl(new NumberReteControl(this.editor, 'depth', '深度'))

        var height = new Rete.Input('height', '高度', PMG.NodesEditor.sockets.number)
        height.addControl(new NumberReteControl(this.editor, 'height', '高度'))

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(width)
            .addInput(depth)
            .addInput(height)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class DrawPrismReteComponent extends Rete.Component {

    constructor() {
        super('绘制棱柱')
    }

    builder(node) {

        var radius = new Rete.Input('radius', '半径', PMG.NodesEditor.sockets.number)
        radius.addControl(new NumberReteControl(this.editor, 'radius', '半径'))

        var height = new Rete.Input('height', '高度', PMG.NodesEditor.sockets.number)
        height.addControl(new NumberReteControl(this.editor, 'height', '高度'))
        
        var sides = new Rete.Input('sides', '边数', PMG.NodesEditor.sockets.number)
        sides.addControl(new NumberReteControl(this.editor, 'sides', '边数'))

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(radius)
            .addInput(height)
            .addInput(sides)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class DrawCylinderReteComponent extends Rete.Component {

    constructor() {
        super('绘制圆柱')
    }

    builder(node) {

        var radius = new Rete.Input('radius', '半径', PMG.NodesEditor.sockets.number)
        radius.addControl(new NumberReteControl(this.editor, 'radius', '半径'))

        var height = new Rete.Input('height', '高度', PMG.NodesEditor.sockets.number)
        height.addControl(new NumberReteControl(this.editor, 'height', '高度'))

        var segments = new Rete.Input('segments', '段数', PMG.NodesEditor.sockets.number)
        segments.addControl(new NumberReteControl(this.editor, 'segments', '段数'))

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(radius)
            .addInput(height)
            .addInput(segments)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class DrawTubeReteComponent extends Rete.Component {

    constructor() {
        super('绘制圆管')
    }

    builder(node) {

        var radius = new Rete.Input('radius', '半径', PMG.NodesEditor.sockets.number)
        radius.addControl(new NumberReteControl(this.editor, 'radius', '半径'))

        var thickness = new Rete.Input('thickness', '厚度', PMG.NodesEditor.sockets.number)
        thickness.addControl(new NumberReteControl(this.editor, 'thickness', '厚度'))

        var height = new Rete.Input('height', '高度', PMG.NodesEditor.sockets.number)
        height.addControl(new NumberReteControl(this.editor, 'height', '高度'))

        var segments = new Rete.Input('segments', '段数', PMG.NodesEditor.sockets.number)
        segments.addControl(new NumberReteControl(this.editor, 'segments', '段数'))

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(radius)
            .addInput(thickness)
            .addInput(height)
            .addInput(segments)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class DrawPyramidReteComponent extends Rete.Component {

    constructor() {
        super('绘制棱锥')
    }

    builder(node) {

        var radius = new Rete.Input('radius', '半径', PMG.NodesEditor.sockets.number)
        radius.addControl(new NumberReteControl(this.editor, 'radius', '半径'))

        var height = new Rete.Input('height', '高度', PMG.NodesEditor.sockets.number)
        height.addControl(new NumberReteControl(this.editor, 'height', '高度'))

        var sides = new Rete.Input('sides', '边数', PMG.NodesEditor.sockets.number)
        sides.addControl(new NumberReteControl(this.editor, 'sides', '边数'))

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(radius)
            .addInput(height)
            .addInput(sides)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class DrawConeReteComponent extends Rete.Component {

    constructor() {
        super('绘制圆锥')
    }

    builder(node) {

        var radius = new Rete.Input('radius', '半径', PMG.NodesEditor.sockets.number)
        radius.addControl(new NumberReteControl(this.editor, 'radius', '半径'))

        var height = new Rete.Input('height', '高度', PMG.NodesEditor.sockets.number)
        height.addControl(new NumberReteControl(this.editor, 'height', '高度'))

        var segments = new Rete.Input('segments', '段数', PMG.NodesEditor.sockets.number)
        segments.addControl(new NumberReteControl(this.editor, 'segments', '段数'))

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(radius)
            .addInput(height)
            .addInput(segments)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class DrawSphereReteComponent extends Rete.Component {

    constructor() {
        super('绘制球体')
    }

    builder(node) {

        var radius = new Rete.Input('radius', '半径', PMG.NodesEditor.sockets.number)
        radius.addControl(new NumberReteControl(this.editor, 'radius', '半径'))

        var segments = new Rete.Input('segments', '段数', PMG.NodesEditor.sockets.number)
        segments.addControl(new NumberReteControl(this.editor, 'segments', '段数'))

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(radius)
            .addInput(segments)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class DrawShapeReteComponent extends Rete.Component {

    constructor() {
        super('Draw shape')
    }

    builder(node) {

        var group = new Rete.Output('groups', 'Group', PMG.NodesEditor.sockets.groups)

        return node
            .addControl(new TextReteControl(this.editor, 'name', 'Name'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class NumberReteComponent extends Rete.Component {

    constructor() {
        super('数量')
    }

    builder(node) {

        var number = new Rete.Output('number', '数量', PMG.NodesEditor.sockets.number)

        return node
            .addControl(new TextReteControl(this.editor, 'label', '标签'))
            .addControl(new NumberReteControl(this.editor, 'number'))
            .addOutput(number)

    }

    worker(node, _inputs, outputs) {
        outputs['number'] = PMG.Utils.isValidNumber(node.data.number) ? node.data.number : 0
    }

}

class AddReteComponent extends Rete.Component {

    constructor() {
        super('加')
    }

    builder(node) {

        var inputNumber1 = new Rete.Input('number1', '数量', PMG.NodesEditor.sockets.number)
        inputNumber1.addControl(new NumberReteControl(this.editor, 'number1'))

        var inputNumber2 = new Rete.Input('number2', '数量', PMG.NodesEditor.sockets.number)
        inputNumber2.addControl(new NumberReteControl(this.editor, 'number2'))

        var outputNumber = new Rete.Output('number', '数量', PMG.NodesEditor.sockets.number)

        return node
            .addInput(inputNumber1)
            .addInput(inputNumber2)
            .addControl(new NumberReteControl(this.editor, 'preview', '', true))
            .addOutput(outputNumber)

    }

    worker(node, inputs, outputs) {

        var number1 = inputs['number1'].length ? inputs['number1'][0] : node.data.number1
        var number2 = inputs['number2'].length ? inputs['number2'][0] : node.data.number2

        if ( PMG.Utils.isValidNumber(number1) ) {
            number1 = parseFloat(number1)
        } else {
            number1 = 0
        }

        if ( PMG.Utils.isValidNumber(number2) ) {
            number2 = parseFloat(number2)
        } else {
            number2 = 0
        }
        
        var sum = number1 + number2
        sum = Math.round((sum + Number.EPSILON) * 1000000) / 1000000
        
        this.editor.nodes.find(n => n.id == node.id).controls.get('preview').setValue(sum)
        outputs['number'] = sum

    }

}

class SubtractReteComponent extends Rete.Component {

    constructor() {
        super('减')
    }

    builder(node) {

        var inputNumber1 = new Rete.Input('number1', '数量', PMG.NodesEditor.sockets.number)
        inputNumber1.addControl(new NumberReteControl(this.editor, 'number1'))

        var inputNumber2 = new Rete.Input('number2', '数量', PMG.NodesEditor.sockets.number)
        inputNumber2.addControl(new NumberReteControl(this.editor, 'number2'))

        var outputNumber = new Rete.Output('number', '数量', PMG.NodesEditor.sockets.number)

        return node
            .addInput(inputNumber1)
            .addInput(inputNumber2)
            .addControl(new NumberReteControl(this.editor, 'preview', '', true))
            .addOutput(outputNumber)

    }

    worker(node, inputs, outputs) {

        var number1 = inputs['number1'].length ? inputs['number1'][0] : node.data.number1
        var number2 = inputs['number2'].length ? inputs['number2'][0] : node.data.number2

        if ( PMG.Utils.isValidNumber(number1) ) {
            number1 = parseFloat(number1)
        } else {
            number1 = 0
        }

        if ( PMG.Utils.isValidNumber(number2) ) {
            number2 = parseFloat(number2)
        } else {
            number2 = 0
        }
        
        var diff = number1 - number2
        diff = Math.round((diff + Number.EPSILON) * 1000000) / 1000000
        
        this.editor.nodes.find(n => n.id == node.id).controls.get('preview').setValue(diff)
        outputs['number'] = diff

    }

}

class MultiplyReteComponent extends Rete.Component {

    constructor() {
        super('乘')
    }

    builder(node) {

        var inputNumber1 = new Rete.Input('number1', '数量', PMG.NodesEditor.sockets.number)
        inputNumber1.addControl(new NumberReteControl(this.editor, 'number1'))

        var inputNumber2 = new Rete.Input('number2', '数量', PMG.NodesEditor.sockets.number)
        inputNumber2.addControl(new NumberReteControl(this.editor, 'number2'))

        var outputNumber = new Rete.Output('number', '数量', PMG.NodesEditor.sockets.number)

        return node
            .addInput(inputNumber1)
            .addInput(inputNumber2)
            .addControl(new NumberReteControl(this.editor, 'preview', '', true))
            .addOutput(outputNumber)

    }

    worker(node, inputs, outputs) {

        var number1 = inputs['number1'].length ? inputs['number1'][0] : node.data.number1
        var number2 = inputs['number2'].length ? inputs['number2'][0] : node.data.number2

        if ( PMG.Utils.isValidNumber(number1) ) {
            number1 = parseFloat(number1)
        } else {
            number1 = 0
        }

        if ( PMG.Utils.isValidNumber(number2) ) {
            number2 = parseFloat(number2)
        } else {
            number2 = 0
        }
        
        var product = number1 * number2
        
        this.editor.nodes.find(n => n.id == node.id).controls.get('preview').setValue(product)
        outputs['number'] = product

    }

}

class DivideReteComponent extends Rete.Component {

    constructor() {
        super('除')
    }

    builder(node) {

        var dividend = new Rete.Input('dividend', '被除数', PMG.NodesEditor.sockets.number)
        dividend.addControl(new NumberReteControl(this.editor, 'dividend', '被除数'))

        var divisor = new Rete.Input('divisor', '除数', PMG.NodesEditor.sockets.number)
        divisor.addControl(new NumberReteControl(this.editor, 'divisor', '除数'))

        var quotient = new Rete.Output('quotient', '商数', PMG.NodesEditor.sockets.number)
        var remainder = new Rete.Output('remainder', '余数', PMG.NodesEditor.sockets.number)

        return node
            .addInput(dividend)
            .addInput(divisor)
            .addControl(new NumberReteControl(this.editor, 'preview', '', true))
            .addOutput(quotient)
            .addOutput(remainder)

    }

    worker(node, inputs, outputs) {

        var dividend = inputs['dividend'].length ? inputs['dividend'][0] : node.data.dividend
        var divisor = inputs['divisor'].length ? inputs['divisor'][0] : node.data.divisor

        if ( PMG.Utils.isValidNumber(dividend) ) {
            dividend = parseFloat(dividend)
        } else {
            dividend = 0
        }

        if ( PMG.Utils.isValidNumber(divisor) ) {
            divisor = parseFloat(divisor)
        } else {
            divisor = 1
        }

        var quotient = dividend / divisor
        var remainder = dividend % divisor
        
        this.editor.nodes.find(n => n.id == node.id).controls.get('preview').setValue(quotient)
        outputs['quotient'] = quotient
        outputs['remainder'] = remainder

    }

}

class CalculateReteComponent extends Rete.Component {

    constructor() {
        super('计算')
    }

    builder(node) {

        var inputA = new Rete.Input('a', '变量 A', PMG.NodesEditor.sockets.number)
        inputA.addControl(new NumberReteControl(this.editor, 'a', '变量 A'))

        var inputB = new Rete.Input('b', '变量 B', PMG.NodesEditor.sockets.number)
        inputB.addControl(new NumberReteControl(this.editor, 'b', '变量 B'))

        var inputC = new Rete.Input('c', '变量 C', PMG.NodesEditor.sockets.number)
        inputC.addControl(new NumberReteControl(this.editor, 'c', '变量 C'))

        var inputD = new Rete.Input('d', '变量 D', PMG.NodesEditor.sockets.number)
        inputD.addControl(new NumberReteControl(this.editor, 'd', '变量 D'))

        var inputE = new Rete.Input('e', '变量 E', PMG.NodesEditor.sockets.number)
        inputE.addControl(new NumberReteControl(this.editor, 'e', '变量 E'))
        
        var inputF = new Rete.Input('f', '变量 F', PMG.NodesEditor.sockets.number)
        inputF.addControl(new NumberReteControl(this.editor, 'f', '变量 F'))

        var inputG = new Rete.Input('g', '变量 G', PMG.NodesEditor.sockets.number)
        inputG.addControl(new NumberReteControl(this.editor, 'g', '变量 G'))

        var inputH = new Rete.Input('h', '变量 H', PMG.NodesEditor.sockets.number)
        inputH.addControl(new NumberReteControl(this.editor, 'h', '变量 H'))

        var inputI = new Rete.Input('i', '变量 I', PMG.NodesEditor.sockets.number)
        inputI.addControl(new NumberReteControl(this.editor, 'i', '变量 I'))

        var inputJ = new Rete.Input('j', '变量 J', PMG.NodesEditor.sockets.number)
        inputJ.addControl(new NumberReteControl(this.editor, 'j', '变量 J'))

        var inputK = new Rete.Input('k', '变量 K', PMG.NodesEditor.sockets.number)
        inputK.addControl(new NumberReteControl(this.editor, 'k', '变量 K'))

        var inputL = new Rete.Input('l', '变量 L', PMG.NodesEditor.sockets.number)
        inputL.addControl(new NumberReteControl(this.editor, 'l', '变量 L'))

        var outputNumber = new Rete.Output('number', '数量', PMG.NodesEditor.sockets.number)

        return node
            .addControl(new TextReteControl(this.editor, 'formula', '公式示例： round(a) * b'))
            .addInput(inputA)
            .addInput(inputB)
            .addInput(inputC)
            .addInput(inputD)
            .addInput(inputE)
            .addInput(inputF)
            .addInput(inputG)
            .addInput(inputH)
            .addInput(inputI)
            .addInput(inputJ)
            .addInput(inputK)
            .addInput(inputL)
            .addOutput(outputNumber)

    }

    worker(node, inputs, outputs) {

        if ( node.data.formula === undefined ) {

            outputs['number'] = 0
            return
            
        }

        var a = inputs['a'].length ? inputs['a'][0] : node.data.a
        var b = inputs['b'].length ? inputs['b'][0] : node.data.b
        var c = inputs['c'].length ? inputs['c'][0] : node.data.c
        var d = inputs['d'].length ? inputs['d'][0] : node.data.d
        var e = inputs['e'].length ? inputs['e'][0] : node.data.e
        var f = inputs['f'].length ? inputs['f'][0] : node.data.f
        var g = inputs['g'].length ? inputs['g'][0] : node.data.g
        var h = inputs['h'].length ? inputs['h'][0] : node.data.h
        var i = inputs['i'].length ? inputs['i'][0] : node.data.i
        var j = inputs['j'].length ? inputs['j'][0] : node.data.j
        var k = inputs['k'].length ? inputs['k'][0] : node.data.k
        var l = inputs['l'].length ? inputs['l'][0] : node.data.l
        var pi = Math.PI

        a = PMG.Utils.isValidNumber(a) ? a : 0
        b = PMG.Utils.isValidNumber(b) ? b : 0
        c = PMG.Utils.isValidNumber(c) ? c : 0
        d = PMG.Utils.isValidNumber(d) ? d : 0
        e = PMG.Utils.isValidNumber(e) ? e : 0
        f = PMG.Utils.isValidNumber(f) ? f : 0
        g = PMG.Utils.isValidNumber(g) ? g : 0
        h = PMG.Utils.isValidNumber(h) ? h : 0
        i = PMG.Utils.isValidNumber(i) ? i : 0
        j = PMG.Utils.isValidNumber(j) ? j : 0
        k = PMG.Utils.isValidNumber(k) ? k : 0
        l = PMG.Utils.isValidNumber(l) ? l : 0

        var fixed_formula = node.data.formula
            .replace(/min/g, 'Math.min')
            .replace(/max/g, 'Math.max')
            .replace(/round/g, 'Math.round')
            .replace(/ceil/g, 'Math.ceil')
            .replace(/floor/g, 'Math.floor')
            .replace(/asinh/g, 'Math.asinh')
            .replace(/asin/g, 'Math.asin')
            .replace(/sin/g, 'Math.sin')
            .replace(/acosh/g, 'Math.acosh')
            .replace(/acos/g, 'Math.acos')
            .replace(/cos/g, 'Math.cos')
            .replace(/atanh/g, 'Math.atanh')
            .replace(/atan/g, 'Math.atan')
            .replace(/tan/g, 'Math.tan')
            .replace(/exp/g, 'Math.exp')
            .replace(/log2/g, 'Math.log2')
            .replace(/log10/g, 'Math.log10')
            .replace(/sqrt/g, 'Math.sqrt')
            .replace(/cbrt/g, 'Math.cbrt')

        outputs['number'] = eval(fixed_formula)

    }

}

class PointReteComponent extends Rete.Component {

    constructor() {
        super('指向')
    }

    builder(node) {

        var inputPointX = new Rete.Input('x', 'X轴', PMG.NodesEditor.sockets.number)
        inputPointX.addControl(new NumberReteControl(this.editor, 'x', 'X轴'))

        var inputPointY = new Rete.Input('y', 'Y轴', PMG.NodesEditor.sockets.number)
        inputPointY.addControl(new NumberReteControl(this.editor, 'y', 'Y轴'))

        var inputPointZ = new Rete.Input('z', 'Z轴', PMG.NodesEditor.sockets.number)
        inputPointZ.addControl(new NumberReteControl(this.editor, 'z', 'Z轴'))

        var outputPoint = new Rete.Output('point', '指向', PMG.NodesEditor.sockets.point)

        return node
            .addInput(inputPointX)
            .addInput(inputPointY)
            .addInput(inputPointZ)
            .addOutput(outputPoint)

    }

    worker(node, _inputs, outputs) {

        var x = PMG.Utils.isValidNumber(node.data.x) ? node.data.x : 0
        var y = PMG.Utils.isValidNumber(node.data.y) ? node.data.y : 0
        var z = PMG.Utils.isValidNumber(node.data.z) ? node.data.z : 0

        outputs['point'] = { x, y, z }

    }
}

class VectorReteComponent extends Rete.Component {

    constructor() {
        super('向量')
    }

    builder(node) {

        var inputVectorX = new Rete.Input('x', 'X轴', PMG.NodesEditor.sockets.number)
        inputVectorX.addControl(new NumberReteControl(this.editor, 'x', 'X轴'))

        var inputVectorY = new Rete.Input('y', 'Y轴', PMG.NodesEditor.sockets.number)
        inputVectorY.addControl(new NumberReteControl(this.editor, 'y', 'Y轴'))

        var inputVectorZ = new Rete.Input('z', 'Z轴', PMG.NodesEditor.sockets.number)
        inputVectorZ.addControl(new NumberReteControl(this.editor, 'z', 'Z轴'))

        var outputVector = new Rete.Output('vector', '向量', PMG.NodesEditor.sockets.vector)

        return node
            .addInput(inputVectorX)
            .addInput(inputVectorY)
            .addInput(inputVectorZ)
            .addOutput(outputVector)

    }

    worker(node, _inputs, outputs) {

        var x = PMG.Utils.isValidNumber(node.data.x) ? node.data.x : 0
        var y = PMG.Utils.isValidNumber(node.data.y) ? node.data.y : 0
        var z = PMG.Utils.isValidNumber(node.data.z) ? node.data.z : 0

        outputs['vector'] = { x, y, z }

    }
}

class IntersectSolidsReteComponent extends Rete.Component {

    constructor() {
        super('实体相交')
    }

    builder(node) {

        var inputGroups1 = new Rete.Input('groups1', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups2 = new Rete.Input('groups2', '群组', PMG.NodesEditor.sockets.groups)

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups1)
            .addInput(inputGroups2)
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class UniteSolidsReteComponent extends Rete.Component {

    constructor() {
        super('实体合并')
    }

    builder(node) {

        var inputGroups1 = new Rete.Input('groups1', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups2 = new Rete.Input('groups2', '群组', PMG.NodesEditor.sockets.groups)

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups1)
            .addInput(inputGroups2)
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class SubtractSolidsReteComponent extends Rete.Component {

    constructor() {
        super('实体减去')
    }

    builder(node) {

        var inputGroups1 = new Rete.Input('groups1', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups2 = new Rete.Input('groups2', '群组', PMG.NodesEditor.sockets.groups)

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups1)
            .addInput(inputGroups2)
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class PushPullReteComponent extends Rete.Component {

    constructor() {
        super('推拉')
    }

    builder(node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)
        var inputDistance = new Rete.Input('distance', '距离', PMG.NodesEditor.sockets.number)
        inputDistance.addControl(new NumberReteControl(this.editor, 'distance', '距离'))
        var inputDirection = new Rete.Input('direction', '方向', PMG.NodesEditor.sockets.vector)

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups)
            .addInput(inputDistance)
            .addControl(new CheckBoxReteControl(this.editor, 'increment_distance', '增量距离'))
            .addInput(inputDirection)
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class MoveReteComponent extends Rete.Component {

    constructor() {
        super('移动')
    }

    builder(node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)
        var inputPoint = new Rete.Input('point', '指向', PMG.NodesEditor.sockets.point)

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups)
            .addInput(inputPoint)
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class RotateReteComponent extends Rete.Component {

    constructor() {
        super('旋转')
    }

    builder(node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)
        var inputCenter = new Rete.Input('center', '中心', PMG.NodesEditor.sockets.point)
        var inputAxis = new Rete.Input('axis', '轴', PMG.NodesEditor.sockets.vector)
        var inputAngle = new Rete.Input('angle', '角度', PMG.NodesEditor.sockets.number)
        inputAngle.addControl(new NumberReteControl(this.editor, 'angle', '角度'))

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups)
            .addInput(inputCenter)
            .addInput(inputAxis)
            .addInput(inputAngle)
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class ScaleReteComponent extends Rete.Component {

    constructor() {
        super('比例')
    }

    builder(node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)
        var inputPoint = new Rete.Input('point', '指向', PMG.NodesEditor.sockets.point)
        var inputXFactor = new Rete.Input('x_factor', 'X 系数', PMG.NodesEditor.sockets.number)
        inputXFactor.addControl(new NumberReteControl(this.editor, 'x_factor', 'X 系数'))
        var inputYFactor = new Rete.Input('y_factor', 'Y 系数', PMG.NodesEditor.sockets.number)
        inputYFactor.addControl(new NumberReteControl(this.editor, 'y_factor', 'Y 系数'))
        var inputZFactor = new Rete.Input('z_factor', 'Z 系数', PMG.NodesEditor.sockets.number)
        inputZFactor.addControl(new NumberReteControl(this.editor, 'z_factor', 'Z 系数'))

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups)
            .addInput(inputPoint)
            .addInput(inputXFactor)
            .addInput(inputYFactor)
            .addInput(inputZFactor)
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class PaintReteComponent extends Rete.Component {

    constructor() {
        super('油漆桶')
    }

    builder(node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups)
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class TagReteComponent extends Rete.Component {

    constructor() {
        super('标签')
    }

    builder(node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups)
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class EraseReteComponent extends Rete.Component {

    constructor() {
        super('擦除')
    }

    builder(node) {

        var groups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(groups)

    }

    worker(_node, _inputs, _outputs) {
    }

}

class CopyReteComponent extends Rete.Component {

    constructor() {
        super('复制')
    }

    builder(node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)
        var inputCopies = new Rete.Input('copies', '副本', PMG.NodesEditor.sockets.number)
        inputCopies.addControl(new NumberReteControl(this.editor, 'copies', '副本'))

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups)
            .addInput(inputCopies)
            .addControl(new CheckBoxReteControl(this.editor, 'output_original', '输出原始'))
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class ConcatenateReteComponent extends Rete.Component {

    constructor() {
        super('连接')
    }

    builder(node) {

        var inputGroups1 = new Rete.Input('groups1', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups2 = new Rete.Input('groups2', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups3 = new Rete.Input('groups3', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups4 = new Rete.Input('groups4', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups5 = new Rete.Input('groups5', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups6 = new Rete.Input('groups6', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups7 = new Rete.Input('groups7', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups8 = new Rete.Input('groups8', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups9 = new Rete.Input('groups9', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups10 = new Rete.Input('groups10', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups11 = new Rete.Input('groups11', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups12 = new Rete.Input('groups12', '群组', PMG.NodesEditor.sockets.groups)

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups1)
            .addInput(inputGroups2)
            .addInput(inputGroups3)
            .addInput(inputGroups4)
            .addInput(inputGroups5)
            .addInput(inputGroups6)
            .addInput(inputGroups7)
            .addInput(inputGroups8)
            .addInput(inputGroups9)
            .addInput(inputGroups10)
            .addInput(inputGroups11)
            .addInput(inputGroups12)
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

class SelectReteComponent extends Rete.Component {

    constructor() {
        super('选择')
    }

    builder(node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)

        var inputA = new Rete.Input('a', '变量 A', PMG.NodesEditor.sockets.number)
        inputA.addControl(new NumberReteControl(this.editor, 'a', '变量 A'))

        var inputB = new Rete.Input('b', '变量 B', PMG.NodesEditor.sockets.number)
        inputB.addControl(new NumberReteControl(this.editor, 'b', '变量 B'))

        var inputC = new Rete.Input('c', '变量 C', PMG.NodesEditor.sockets.number)
        inputC.addControl(new NumberReteControl(this.editor, 'c', '变量 C'))

        var inputD = new Rete.Input('d', '变量 D', PMG.NodesEditor.sockets.number)
        inputD.addControl(new NumberReteControl(this.editor, 'd', '变量 D'))

        var inputE = new Rete.Input('e', '变量 E', PMG.NodesEditor.sockets.number)
        inputE.addControl(new NumberReteControl(this.editor, 'e', '变量 E'))
        
        var inputF = new Rete.Input('f', '变量 F', PMG.NodesEditor.sockets.number)
        inputF.addControl(new NumberReteControl(this.editor, 'f', '变量 F'))

        var inputG = new Rete.Input('g', '变量 G', PMG.NodesEditor.sockets.number)
        inputG.addControl(new NumberReteControl(this.editor, 'g', '变量 G'))

        var inputH = new Rete.Input('h', '变量 H', PMG.NodesEditor.sockets.number)
        inputH.addControl(new NumberReteControl(this.editor, 'h', '变量 H'))

        var inputI = new Rete.Input('i', '变量 I', PMG.NodesEditor.sockets.number)
        inputI.addControl(new NumberReteControl(this.editor, 'i', '变量 I'))

        var inputJ = new Rete.Input('j', '变量 J', PMG.NodesEditor.sockets.number)
        inputJ.addControl(new NumberReteControl(this.editor, 'j', '变量 J'))

        var inputK = new Rete.Input('k', '变量 K', PMG.NodesEditor.sockets.number)
        inputK.addControl(new NumberReteControl(this.editor, 'k', '变量 K'))

        var inputL = new Rete.Input('l', '变量 L', PMG.NodesEditor.sockets.number)
        inputL.addControl(new NumberReteControl(this.editor, 'l', '变量 L'))

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)
        
        var outputNotGroups = new Rete.Output('not_groups', '不匹配的群组', PMG.NodesEditor.sockets.groups)

        return node
            .addControl(new TextReteControl(this.editor, 'query', '查询示例：first or even'))
            .addInput(inputGroups)
            .addInput(inputA)
            .addInput(inputB)
            .addInput(inputC)
            .addInput(inputD)
            .addInput(inputE)
            .addInput(inputF)
            .addInput(inputG)
            .addInput(inputH)
            .addInput(inputI)
            .addInput(inputJ)
            .addInput(inputK)
            .addInput(inputL)
            .addOutput(outputGroups)
            .addOutput(outputNotGroups)

    }

    worker(_node, _inputs, outputs) {

        outputs['groups'] = []
        outputs['not_groups'] = []

    }

}

class MakeGroupReteComponent extends Rete.Component {

    constructor() {
        super('群组')
    }

    builder(node) {

        var inputGroups1 = new Rete.Input('groups1', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups2 = new Rete.Input('groups2', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups3 = new Rete.Input('groups3', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups4 = new Rete.Input('groups4', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups5 = new Rete.Input('groups5', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups6 = new Rete.Input('groups6', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups7 = new Rete.Input('groups7', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups8 = new Rete.Input('groups8', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups9 = new Rete.Input('groups9', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups10 = new Rete.Input('groups10', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups11 = new Rete.Input('groups11', '群组', PMG.NodesEditor.sockets.groups)
        var inputGroups12 = new Rete.Input('groups12', '群组', PMG.NodesEditor.sockets.groups)

        var outputGroup = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups1)
            .addInput(inputGroups2)
            .addInput(inputGroups3)
            .addInput(inputGroups4)
            .addInput(inputGroups5)
            .addInput(inputGroups6)
            .addInput(inputGroups7)
            .addInput(inputGroups8)
            .addInput(inputGroups9)
            .addInput(inputGroups10)
            .addInput(inputGroups11)
            .addInput(inputGroups12)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(outputGroup)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = []
    }

}

PMG.codeName = 'ParametricModeling'

PMG.NodesEditor.initializeEditor = () => {

    PMG.NodesEditor.editor = new Rete.NodeEditor(
        PMG.codeName + '@' + PMGNodesEditorSchemaVersion,
        document.querySelector('#pmg-nodes-editor')
    )

    PMG.NodesEditor.editor.use(VueRenderPlugin.default)
    PMG.NodesEditor.editor.use(ConnectionPlugin.default)
    PMG.NodesEditor.editor.use(MinimapPlugin.default)

}

PMG.NodesEditor.initializeEngine = () => {

    PMG.NodesEditor.engine = new Rete.Engine(
        PMG.codeName + '@' + PMGNodesEditorSchemaVersion
    )

}

PMG.NodesEditor.initializeComponents = () => {

    PMG.NodesEditor.components = {

        "绘制盒子": new DrawBoxReteComponent(),
        "绘制棱柱": new DrawPrismReteComponent(),
        "绘制圆柱": new DrawCylinderReteComponent(),
        "绘制圆管": new DrawTubeReteComponent(),
        "绘制棱锥": new DrawPyramidReteComponent(),
        "绘制圆锥": new DrawConeReteComponent(),
        "绘制球体": new DrawSphereReteComponent(),
        "Draw shape": new DrawShapeReteComponent(),
        "数量": new NumberReteComponent(),
        "加": new AddReteComponent(),
        "减": new SubtractReteComponent(),
        "乘": new MultiplyReteComponent(),
        "除": new DivideReteComponent(),
        "计算": new CalculateReteComponent(),
        "指向": new PointReteComponent(),
        "向量": new VectorReteComponent(),
        "实体相交": new IntersectSolidsReteComponent(),
        "实体合并": new UniteSolidsReteComponent(),
        "实体减去": new SubtractSolidsReteComponent(),
        "推拉": new PushPullReteComponent(),
        "移动": new MoveReteComponent(),
        "旋转": new RotateReteComponent(),
        "比例": new ScaleReteComponent(),
        "油漆桶": new PaintReteComponent(),
        "标签": new TagReteComponent(),
        "擦除": new EraseReteComponent(),
        "复制": new CopyReteComponent(),
        "连接": new ConcatenateReteComponent(),
        "选择": new SelectReteComponent(),
        "群组": new MakeGroupReteComponent()

    }

}

PMG.NodesEditor.registerComponents = () => {

    for (var nodeName in PMG.NodesEditor.components) {

        var component = PMG.NodesEditor.components[nodeName]

        PMG.NodesEditor.editor.register(component)
        PMG.NodesEditor.engine.register(component)

    }

}

PMG.NodesEditor.loadToolbarIcons = () => {
    
    document.querySelectorAll('.toolbar .node-icon').forEach(toolbarNodeIcon => {

        toolbarNodeIcon.src = PMGNodesEditorIcons['nodes'][toolbarNodeIcon.dataset.nodeName]['path']
        toolbarNodeIcon.title = toolbarNodeIcon.dataset.nodeName

    })

    var toolbarHelpIcon = document.querySelector('.toolbar .help-icon')

    toolbarHelpIcon.src = PMGNodesEditorIcons['help']['path']
    toolbarHelpIcon.title = PMGNodesEditorIcons['help']['title']

    if ( SketchUpVersion < 21 ) {
        new Drooltip({
            element: '.toolbar .node-icon, .toolbar .help-icon',
            position: 'bottom',
            background: '#fff',
            color: '#000',
            animation: 'fade'
        })
    }

}

PMG.NodesEditor.schemaIsExportable = false

PMG.NodesEditor.exportModelSchema = redraw => {

    if ( PMG.NodesEditor.schemaIsExportable ) {
        sketchup.exportModelSchema(JSON.stringify(PMG.NodesEditor.editor.toJSON()), redraw)
    }

}

PMG.NodesEditor.addNode = (nodeName, nodeData) => {

    var component = PMG.NodesEditor.components[nodeName]
    var mouse = PMG.NodesEditor.editor.view.area.mouse

    component.createNode(nodeData).then(node => {

        node.position = [mouse.x, mouse.y]
        PMG.NodesEditor.editor.addNode(node)

        PMG.NodesEditor.nodeBeingAdded = node
        
    })

}

PMG.NodesEditor.resizeEditorView = () => {
    document.querySelector('#pmg-nodes-editor').style.height = window.innerHeight + 'px'
}

PMG.NodesEditor.addEventListeners = () => {

    PMG.NodesEditor.editor.on('process nodecreated noderemoved connectioncreated connectionremoved', () => {
        
        PMG.NodesEditor.engine.abort().then(() => {
            PMG.NodesEditor.engine.process(PMG.NodesEditor.editor.toJSON())
        })
        
    })

    PMG.NodesEditor.editor.on('nodecreated noderemoved connectioncreated connectionremoved', () => {
        PMG.NodesEditor.exportModelSchema(true)
    })

    PMG.NodesEditor.editor.on('nodedragged', () => {
        PMG.NodesEditor.exportModelSchema(false)
    })

    PMG.NodesEditor.editor.on('nodecreated', (node) => {

        var nodeElement = node.vueContext.$el

        nodeElement.setAttribute('data-node-id', node.id)

        new ContextMenu('.node[data-node-id="' + node.id + '"] *', [
            {
                name: '删除此节点',
                fn: () => { PMG.NodesEditor.editor.removeNode(node) }
            }
        ])

        if ( SketchUpVersion < 21 ) {
            new Drooltip({
                element: '.node[data-node-id="' + node.id + '"] .socket',
                position: 'bottom',
                background: '#fff',
                color: '#000',
                animation: 'fade'
            })
        }

        nodeElement.querySelectorAll('input[type="number"]').forEach(nodeNumberInputElement => {

            nodeNumberInputElement.addEventListener('input', _event => {
                PMG.NodesEditor.exportModelSchema(true)
            })

        })

        nodeElement.querySelectorAll('input[type="text"]').forEach(nodeTextInputElement => {

            nodeTextInputElement.addEventListener('input', _event => {
                PMG.NodesEditor.exportModelSchema(true)
            })

        })

        nodeElement.querySelectorAll('input[type="checkbox"]').forEach(nodeCheckBoxInputElement => {

            nodeCheckBoxInputElement.addEventListener('change', _event => {
                PMG.NodesEditor.exportModelSchema(true)
            })

        })

        nodeElement.querySelectorAll('select').forEach(nodeSelectElement => {

            nodeSelectElement.addEventListener('change', _event => {
                PMG.NodesEditor.exportModelSchema(true)
            })

        })

        var nodeTitleElement = nodeElement.querySelector('.title')

        var nodeTitleGradient = 'linear-gradient(0deg,hsla(0,0%,100%,.05) 0,hsla(0,0%,100%,.05)' +
            ' 40%,hsla(0,0%,100%,.19)),radial-gradient(70% 40px at center,' +
            PMGNodesEditorIcons['nodes'][node.name]['color'] + ' 0,rgba(0,0,0,0) 60%)'

        nodeTitleElement.style.backgroundImage = nodeTitleGradient

        var nodeTitleIconElement = document.createElement('img')

        nodeTitleIconElement.src = PMGNodesEditorIcons['nodes'][node.name]['path']
        nodeTitleIconElement.className = 'icon'
        nodeTitleElement.appendChild(nodeTitleIconElement)

    })

    PMG.NodesEditor.editor.on('zoom', ({ source }) => {
        return source !== 'dblclick'
    })

    PMG.NodesEditor.editor.on('mousemove', () => {

        if ( PMG.NodesEditor.nodeBeingAdded === undefined ) {
            return
        }

        var mouse = PMG.NodesEditor.editor.view.area.mouse

        PMG.NodesEditor.editor.view.nodes
            .get(PMG.NodesEditor.nodeBeingAdded)
            .translate(mouse.x, mouse.y)

    })

    window.addEventListener('click', event => {

        if ( event.target.className === 'node-icon' ) {
            return
        }

        PMG.NodesEditor.nodeBeingAdded = undefined

    })

    document.querySelectorAll('.toolbar .node-icon').forEach(toolbarNodeIcon => {

        toolbarNodeIcon.addEventListener('click', event => {
            PMG.NodesEditor.addNode(event.currentTarget.dataset.nodeName, {})
        })

    })

    document.querySelector('.toolbar .help-icon').addEventListener('click', _event => {
        sketchup.accessOnlineHelp()
    })

    window.addEventListener('resize', _event => {
        PMG.NodesEditor.resizeEditorView()
    })

}

PMG.NodesEditor.importModelSchema = () => {

    PMG.NodesEditor.engine.abort().then(() => {

        PMG.NodesEditor.editor.fromJSON(PMGNodesEditorSchema).then(() => {

            PMG.NodesEditor.engine.process(PMG.NodesEditor.editor.toJSON()).then(() => {
                PMG.NodesEditor.schemaIsExportable = true
            })

        })
        
    })

}

PMG.NodesEditor.showOrHideMinimap = () => {

    document.querySelector('.minimap').classList.toggle('displayed')
    PMG.NodesEditor.editor.trigger('zoomed')

}

PMG.NodesEditor.setGlobalContextMenu = () => {

    var contextMenuOptions = [
        {
            name: '导入 schema 文件',
            fn: () => { sketchup.importSchemaFromFile() }
        },
        {
            name: '导出为 schema 文件',
            fn: () => { sketchup.exportSchemaToFile() }
        },
        {
            name: '冻结参数化实体',
            fn: () => { sketchup.freezeParametricEntities() }
        },
        {
            name: '显示或隐藏小地图',
            fn: () => { PMG.NodesEditor.showOrHideMinimap() }
        },
        {
            name: '删除所有节点',
            fn: () => { PMG.NodesEditor.editor.clear() }
        }
    ]

    new ContextMenu('#pmg-nodes-editor', contextMenuOptions)

}

PMG.NodesEditor.tagNodesAsValid = () => {

    document.querySelectorAll('.node').forEach(nodeElement => {
        nodeElement.classList.remove('invalid')
    })

}

PMG.NodesEditor.tagNodeAsInvalid = nodeId => {

    var nodeElement = document.querySelector('.node[data-node-id="' + nodeId + '"]')
    nodeElement.classList.add('invalid')

}

document.addEventListener('DOMContentLoaded', _event => {

    PMG.NodesEditor.initializeEditor()
    PMG.NodesEditor.initializeEngine()
    PMG.NodesEditor.initializeComponents()
    PMG.NodesEditor.registerComponents()
    PMG.NodesEditor.loadToolbarIcons()
    PMG.NodesEditor.addEventListeners()
    PMG.NodesEditor.importModelSchema()
    PMG.NodesEditor.setGlobalContextMenu()
    PMG.NodesEditor.resizeEditorView()

})
